/******************************************************
*
*	Copyright by Kamil Szatkowski "Netrix"
*
*	This code is distributed as Public Domain
*
******************************************************/
#include "GraphColoring.hpp"

#define _SECURE_SCL 0
#define _SECURE_SCL_THROWS 0
#include <stdio.h>
#include <string>
#include <algorithm>
#include <assert.h>
#include <string.h>
#include <set>
using namespace std;

namespace
{
	const uint GENERATIONS_COUNT = 3;
}

bool GraphColoring::LoadGraph(const char* szGraphFile)
{
	FILE* pFile = fopen(szGraphFile, "r");
	if(pFile == 0)
	{
		return true;
	}

	char buffer[2048];
	uint iEdge = 0;

	// Loading file
	while(!feof(pFile))
	{
		char type = 0;
		fscanf(pFile, "%c", &type);

		if(type == 'c')
		{
			fgets(buffer, sizeof(buffer), pFile);
			printf("%s\n", buffer);
		}
		else if(type == 'p')
		{
			assert(m_aEdges.empty() == true);
			assert(m_aFields.empty() == true);

			uint uFieldsCount,
				 uEdgesCount;
			fscanf(pFile, "%s %d %d\n", buffer, &uFieldsCount, &uEdgesCount);

			m_aFields.resize(uFieldsCount);
			m_aEdges.resize(uEdgesCount);
			
			for(uint i = 0; i < uFieldsCount; ++i)
			{
				m_aFields[i].uEdgesCount = 0;
			}
		}
		else if(type == 'e')
		{
			assert(m_aEdges.empty() != true);
			assert(m_aFields.empty() != true);

			uint a, b;
			fscanf(pFile, "%d %d\n", &a, &b);

			m_aEdges[iEdge].uField[0] = a - 1;
			m_aEdges[iEdge].uField[1] = b - 1;
			++iEdge;
			++m_aFields[a - 1].uEdgesCount;
		}
		else
		{
			fgets(buffer, sizeof(buffer), pFile);
		}
	}
	fclose(pFile);

	//  Filling m_aFields
	for(uint iField = 0; iField < m_aFields.size(); ++iField)
	{
		m_aFields[iField].puEdges = new uint[m_aFields[iField].uEdgesCount];
		m_aFields[iField].uEdgesCount = 0;
		m_aFields[iField].uIndex = iField;

		for(uint iEdge = 0; iEdge < m_aEdges.size(); ++iEdge)
		{
			if(m_aEdges[iEdge].uField[0] == iField)
			{
				m_aFields[iField].puEdges[m_aFields[iField].uEdgesCount] = iEdge;
				++m_aFields[iField].uEdgesCount;
			}
		}
	}

	// Creating and zeroing history
	m_aHistory = new uint[GENERATIONS_COUNT * m_uGenerationSize * m_aFields.size()];
	memset(m_aHistory, 0, GENERATIONS_COUNT * m_uGenerationSize * m_aFields.size() * sizeof(uint));

	m_aHistoryConflicts = new uint[GENERATIONS_COUNT * m_uGenerationSize];
	memset(m_aHistoryConflicts, 0xffffffff, GENERATIONS_COUNT * m_uGenerationSize * sizeof(uint));

	uint uGenerationOffset = m_uGenerationSize * m_aFields.size();

	m_pPrevGeneration = m_aHistory;
	m_pActualGeneration = m_aHistory + uGenerationOffset;
	m_pBestGeneration = m_aHistory + 2 * uGenerationOffset;
	m_pPrevGenerationConflicts = m_aHistoryConflicts;
	m_pActualGenerationConflicts = m_aHistoryConflicts + m_uGenerationSize;
	m_pBestGenerationConflicts = m_aHistoryConflicts + 2 * m_uGenerationSize;

	srand((uint)__rdtsc());	// pretty random seed

	return false;
}

void GraphColoring::ReleaseGraph()
{
	delete[] m_aHistory;
	delete[] m_aHistoryConflicts;

	for(uint i = 0; i < m_aFields.size(); ++i)
	{
		delete[] m_aFields[i].puEdges;
	}
}

void GraphColoring::ResetGraph(uint uColorsCount)
{
	m_uColorsCount = uColorsCount;
	memset(m_aHistory, 0, GENERATIONS_COUNT * m_uGenerationSize * m_aFields.size() * sizeof(uint));
	memset(m_aHistoryConflicts, 0xffffffff, GENERATIONS_COUNT * m_uGenerationSize * sizeof(uint));

	srand((uint)__rdtsc());	// pretty random seed
}

void GraphColoring::InitPopulation()
{
	m_uActualGeneration = 0;
	m_uBestIndividual[0] = 0;
	m_uBestIndividual[1] = 0;

	for(uint iInd = 0; iInd < m_uGenerationSize; ++iInd)
	{
		uint* puColors = GetIndividualOffset(m_pActualGeneration, iInd);

		// Initializating first individual
		for(uint iField = 0; iField < m_aFields.size(); ++iField)
		{
			puColors[iField] = rand() % m_uColorsCount;
		}
	}

	EvaluateGeneration(m_pActualGeneration, m_pActualGenerationConflicts, m_uActualGeneration);

	// Initiating first generation as best
	memcpy(m_pBestGeneration, m_pActualGeneration, m_uGenerationSize * m_aFields.size() * sizeof(uint));
	memcpy(m_pBestGenerationConflicts, m_pActualGenerationConflicts, m_uGenerationSize * sizeof(uint));
}

void GraphColoring::EvaluateGeneration(uint* pGeneration, uint* pGenerationConflicts, uint uGeneration)
{
	bool bNewBest = false;

	uint uBestValue = *GetIndividualConflictsOffset(m_pBestGenerationConflicts, m_uBestIndividual[1]);

	for(uint iInd = 0; iInd < m_uGenerationSize; ++iInd)
	{
		uint* puColors = GetIndividualOffset(pGeneration, iInd);
		uint* puConflicts = GetIndividualConflictsOffset(pGenerationConflicts, iInd);

		*puConflicts = EvaluateConflicts(puColors);

		if(uBestValue > *puConflicts)
		{
			// Making copy of whole generation, for curiosity
			if(m_uBestIndividual[0] != uGeneration)
			{
				bNewBest = true;
			}

			uBestValue = *puConflicts;
			m_uBestIndividual[0] = uGeneration;
			m_uBestIndividual[1] = iInd;
		}
	}

	if(bNewBest)
	{
		memcpy(m_pBestGeneration, pGeneration, m_uGenerationSize * m_aFields.size() * sizeof(uint));
		memcpy(m_pBestGenerationConflicts, pGenerationConflicts, m_uGenerationSize * sizeof(uint));
	}
}

uint GraphColoring::EvaluateConflicts(uint* uIndividualOffset)
{
	uint uConflicts = 0;

	for(uint i = 0; i < m_aEdges.size(); ++i)
	{
		uint a = m_aEdges[i].uField[0];
		uint b = m_aEdges[i].uField[1];

		if(uIndividualOffset[a] == uIndividualOffset[b])
		{
			uConflicts++;
		}
	}

	return uConflicts;
}

void GraphColoring::SwapGenerations()
{
	uint* pTemp = m_pPrevGeneration;
	m_pPrevGeneration = m_pActualGeneration;
	m_pActualGeneration = pTemp;

	pTemp = m_pPrevGenerationConflicts;
	m_pPrevGenerationConflicts = m_pActualGenerationConflicts;
	m_pActualGenerationConflicts = pTemp;
}

void GraphColoring::IterateGeneration()
{
	SwapGenerations();

	for(uint i = 0; i < m_uGenerationSize; ++i)
	{
		uint uIndividual = (*m_cbSelectIndividual)(m_pPrevGenerationConflicts, m_uGenerationSize, MAX_UINT);
		uint* pIndividual = GetIndividualOffset(m_pPrevGeneration, uIndividual);
		uint* pNewIndividual = GetIndividualOffset(m_pActualGeneration, i);
		
		if(CheckProbability(m_dCrossingProbability))
		{
			// Needed 2nd Individual
			uint uSecond = (*m_cbSelectIndividual)(m_pPrevGenerationConflicts, m_uGenerationSize, uIndividual);
			uint* pSecond = GetIndividualOffset(m_pPrevGeneration, uSecond);
			uint* pNewSecond = GetIndividualOffset(m_pActualGeneration, i + 1);

			(*m_cbCrossoverIndividuals)(pIndividual, pSecond, m_aFields.size(), pNewIndividual, pNewSecond);

			if(CheckProbability(m_dMutatationProbability))
			{
				(*m_cbMutateIndividual)(pNewSecond, m_aFields.size(), m_uColorsCount);
			}
			++i;
		}
		else
		{
			memcpy(pNewIndividual, pIndividual, m_aFields.size() * sizeof(uint));
		}

		if(CheckProbability(m_dMutatationProbability))
		{
			(*m_cbMutateIndividual)(pNewIndividual, m_aFields.size(), m_uColorsCount);
		}
	}

	EvaluateGeneration(m_pActualGeneration, m_pActualGenerationConflicts, ++m_uActualGeneration);
}

bool GraphColoring::CheckProbability(double dCoeff)
{
	const int ACCURACY = 10000000;
	return (rand() % ACCURACY) < (int)(dCoeff * (double)ACCURACY);
}